# -*- coding: utf-8 -*-
"""
Created on Tue Aug 18 2020

Last edited on Wed May 5 2022

@author: beaufils

Script to format the EORA tables.
Please update the relative folder paths eora_path and save_path.
eora_path should point toward the folder containing the original EORA tables.
save_path points toward the folder in which formatted table will be stored
"""


import os
import numpy as np
import pandas as pd

eora_path = os.path.join('..','data','EORA','raw')
save_path = os.path.join('..','data','EORA','formatted')

def load_label():
    lab = pd.read_csv(os.path.join(eora_path,'labels','labels_T.txt'),
                      sep='\t',header=None).to_numpy()
    extract = lab[:4914,[1,3]]
    return extract

def load_raw_EORA(year,countries=189,sectors=26):
    t = np.loadtxt(os.path.join(eora_path,'Eora26_'+str(year)+'_bp_T.txt'),
                   delimiter = '\t')
    y = np.loadtxt(os.path.join(eora_path,'Eora26_'+str(year)+'_bp_FD.txt'),
                   delimiter = '\t')
    v = np.loadtxt(os.path.join(eora_path,'Eora26_'+str(year)+'_bp_VA.txt'),
                   delimiter = '\t')
    return t[:26*countries,:26*countries],y[:26*countries,:],v[:,:26*countries]

def basic_eora(year,countries=189,sectors=26):
    '''
    Saves a copy of the EORA tables.
    Final demand categories are aggregated.
    Negative changes in stocks and inventories are neglected.
    1 are set in the diagonal of empty sectors, in the diagonal of final demand
    and in value added to avoid numerical issues when computing 
    the Leontief inverse.
    
    Formatted tables are saved in the "save_path" folder.
    
    Parameters
    ----------
    year : int
        Year to treat.
    countries : int, optional
        Number of countries in the initial table. The default is 189.
    sectors : int, optional
        Number of sectors per country. The default is 26.
    
    Returns
    -------
    None.

    '''
    t,y,v = load_raw_EORA(year)
    
    #Associate formatted labels with EORA source labels
    label = load_label()
    clab=[]
    for i in range(189):
        clab.append(label[i*26,0])    
    
    print('Formatting value added and final demand')
    y[y<0] = 0  #Set negative changes in stocks and inventories to 0 to avoid
                #numerical errors
    y1 = np.zeros((countries*sectors,countries))
    
    for c in range(countries):
       y1[:,c] = np.sum(y[:,c*6:(c+1)*6],axis=1)
    v = np.sum(v,axis=0) # Sum all components to value added

    xo = np.sum(t,axis=1) + np.sum(y,axis=1)
    xi = np.sum(t,axis=0) + v
    for i,a in enumerate(xo):
        if a!=0 and xi[i] ==0:
            t[:,i],v[i] = t[i,:],np.sum(y[i,:])
            print('Empty inputs at index '+ str(i))
        if a==0 and xi[i]==0:
            t[i,i],y1[i,i//26],v[i] = 1,1,1
            print('Empty sector at index '+str(i))
    for i,a in enumerate(xi):
        if a!=0 and xo[i] ==0:
            t[i,:],y1[i,i//26] = t[:,i],v[i]
            print('Empty outputs at index '+ str(i))
                
    
    print('Saving the tables')
    np.savetxt(os.path.join(save_path,
                            'Eora26_clean_'+str(year)+'_bp_T.csv'),
               t,delimiter=',')
    np.savetxt(os.path.join(save_path,
                            'Eora26_clean_'+str(year)+'_bp_FD.csv'),
               y1,delimiter=',')
    np.savetxt(os.path.join(save_path,
                            'Eora26_clean_'+str(year)+'_bp_VA.csv'),
               v,delimiter=',')
    np.savetxt(os.path.join(save_path,
                            'country_labels.csv'),
               clab,delimiter=',',fmt='%s')
    np.savetxt(os.path.join(save_path,
                            'full_labels.csv'),
               label,delimiter=',',fmt='%s')
    
basic_eora(2015)