# -*- coding: utf-8 -*-
"""
Created on Wed Jun 19 2019

Last updated on Wed May 04 2022

@author: Timothé Beaufils

Integrates emissions indicators in EORA using PRIMAP hist and PRIMAP crf

V1.0

Please update the path descriptions consistently
"""

import numpy as np
import io
import csv
import os
import pandas as pd

#Path for source files
source_path = os.path.join('..','data')
eora_path = os.path.join(source_path,'EORA')
crf_path = os.path.join(source_path ,'PRIMAP',
                        'Guetschow-et-al-2020-PRIMAP-crf_2020-v1.csv')
hist_path = os.path.join(source_path,'PRIMAP',
                         'PRIMAP-hist_v2.2_19-Jan-2021.csv')
mapping_path = os.path.join(source_path,'mapping.csv')
output_path = os.path.join(source_path,'EORA')

cat_relabel =[['IPC0'],['IPCM0EL']]


def load_eora(year,fd_as_array=False,version='formatted'):
    """
    Main function
    Returns a formatted version of EORA 26

    Parameters
    ----------
    year : int
        Year to load.
    fd_as_array : bool, optional
        Whether to aggregate domestic final demand and foreign final demand. 
        The default is False.
    version : string, optional
        Name of the data version to load. The default is 'formatted'.

    Returns
    -------
    t : 2D array
        Inter-industry matrix.
    y : 1-2D array
        Final demand matrix (or array if fd_as_array is 'True').
    c : int
        Number of countries covered.
    s : int
        Number of sectors covered.
    """
    t = np.loadtxt(os.path.join(eora_path,version,
                                'Eora26_clean_'+str(year)+'_bp_T.csv'),
                   delimiter=',')
    y = np.loadtxt(os.path.join(eora_path,version,
                                'Eora26_clean_'+str(year)+'_bp_FD.csv'),
                   delimiter=',')
    s=26
    c = int(len(t)/s)
    if fd_as_array:
        y= np.sum(y,axis=1)
    return t,y,c,s

def get_go(year,version):
    """
    Extracts the gross output and total final demand per sector from EORA

    Parameters
    ----------
    year : int
        Base year for the EORA table.
    version : string
        Data version for the EORA database.

    Returns
    -------
    list of list of floats
        First list holds gross output per sector.
        Second list holds final demand per sector (domestic + foreign).
    """

    t,y,c,s = load_eora(year,version=version)
    go = np.sum(t,axis = 1)+np.sum(y,axis = 1)
    fdsum = np.sum(y,axis=0)
    return [go,fdsum]

def load_countries(version):
    """
    Import Ordered list of A3 country codes from EORA

    Parameters
    ----------
    version : string
        Version of the EORA database.

    Returns
    -------
    list of ints
        List of A3 countries codes in EORA.

    """
    return np.loadtxt(os.path.join(eora_path,version,'country_labels.csv'),
                      delimiter = ',',dtype=str).tolist()

def load_crf(year,countries,cat_crf,entities):
    """
    Extract data from PRIMAP-crf for a given year and given GHGs.

    Parameters
    ----------
    year : int
        Year data to consider.
    countries : list of strings
        List of A3 country codes for countries covered in the MRIOT.
    cat_crf : list of strings
        Name of GHGs to extract.

    Returns
    -------
    list of arrays
        First array holds labels (country,category,GHG).
        Second array holds values.
    """
    crf_file = io.open(crf_path,"r",encoding = 'utf-8-sig')
    reader = csv.reader(crf_file, delimiter = ',')
    labels = []
    values = []
    for row in reader:
        if (row[1] in countries and row[2] in cat_crf and 
            row[3] in entities and row[year-1981]!='NaN'):
            labels.append([row[1],row[2],row[3]])
            values.append(row[year-1981])
    return [np.array(labels,dtype=str),np.array(values,dtype=float)]

def load_hist(year,countries,cat_crf,entities):
    """
    Extract data from PRIMAP-hist for a given year and given GHGs.

    Parameters
    ----------
    year : int
        Year data to consider.
    countries : list of strings
        List of A3 country codes for countries covered in the MRIOT.
    cat_crf : list of strings
        Name of GHGs to extract.

    Returns
    -------
    list of arrays
        First array holds labels (country,category,GHG).
        Second array holds values.
    """
    primap_file = io.open(hist_path,'r',encoding = 'utf-8-sig')
    reader = csv.reader(primap_file, delimiter = ',')
    labels = []
    values = []
    for row in reader:
        if (row[0] == 'HISTTP' and row[1] in countries and 
            row[2] in cat_crf and row[3] in entities and 
            row[year-1845]!='NaN'):
            labels.append([row[1],row[2],row[3]])
            values.append(row[year-1845])
    return [np.array(labels,dtype=str),np.array(values,dtype=float)]

def load_mapping():
    """
    Load IPCC emission categories and the corresponding EORA sectors.

    Returns
    -------
    cat : list of strings
        Label of IPCC categories.
    mapping : list of ints
        Index of corresponding EORA sector.
    parents : list of ints
        List of index of parent categories.
    childs : list of list of ints
        Lists of indexes of children categories.
        Index of list of children refers to index in parents' list.

    """
    source = np.loadtxt(mapping_path,delimiter=',',dtype=str)
    cat = source[0].tolist()
    mapping = source[1]
    hierarchy = source[2]
    parents = []
    childs = []
    for i,x in enumerate(hierarchy):
        if x != '':
            if int(x) not in parents:
                parents.append(int(x))
            if i==0 or hierarchy[i-1]!= x:
                childs.append([i])
            else:
                childs[-1].append(i)
    return cat,mapping,parents,childs

def check_consistency(mother,child):
    """
    Test function
    Checks the sum of subitems (child) equals the aggregated value (mother)
    Returns the ratio between childs and mother

    Parameters
    ----------
    mother : float
        Value of the parent item.
    child : list of floats
        Values of the children items.

    Returns
    -------
    float
        Returns ratio between mother and child.
    """
    if np.sum(child)!=0:
        return mother/np.sum(child)
    else:
        if mother==0:
            return 1
        else:
            return 0
    
def intensity_factor(year,crf,concordance,cat_crf,countries,go,
                     entities,min_sources=0):
    """
    Computes average emissions intensity coefficient for end-categories.
    End-categories are defined are emission categories associated with 
    a single EORA sector.
    Emission intensity coefficients are computed using PRIMAP-crf emission data
    and EORA economic output values.

    Parameters
    ----------
    year : int
        Year to be covered.
    crf : list of arrays
        Data from PRIMAP-crf, as extracted from load_crf.
    concordance : list of ints
        Mapping table from emission categories to EORA sectors.
    cat_crf : list of strings
        List of IPCC emission categories labels.
    countries : list of string
        List of A3 country codes for countries covered in EORA.
    go : list of float
        Economic output per sector.
    min_sources : int, optional
        Minimal number of sources to create an intensity coefficient. 
        The default is 0.

    Returns
    -------
    intensity : 2D matrix
        Emission intensity coefficient for each GHG and each category.
    sources : 2D matrix
        Number of countries used for determining the emission intensity coeff.

    """
    intensity = np.zeros([len(entities),len(cat_crf)])
    sources = np.zeros([len(entities),len(cat_crf)])
    for rowid,entity in enumerate(entities):
        for colid, category in enumerate(cat_crf):
            countries_name = []
            values = 0
            source = 0
            if not concordance[colid] == '': 
                #If the category is associated with an EORA sector
                sectorindex = int(concordance[colid])
                indices = [(label[1] == category and 
                            label[2] == entity) for label in crf[0]]
                            #List all countries for which data is reported
                source = np.sum(indices)
                values = np.sum(crf[1][indices])
                countries_name = crf[0][indices][:,0]
                sources[rowid,colid] = source
                if source>min_sources: 
                    #If enough sources are available 
                    #computes the average factor of the category
                    index_countries = []
                    for country in countries_name:
                        index_countries.append(countries.index(country))
                    if sectorindex == 26: 
                        #Distinguishes between final demand 
                        #and intermediate sectors
                        intensity[rowid,colid] = values/(np.sum([go[1][i] 
                                                    for i in index_countries]))
                    else:
                        intensity[rowid,colid] = \
                            values/(np.sum([go[0][i*26 + sectorindex] 
                                            for i in index_countries]))
    return intensity,sources

def order_primap(data, countries,cat_crf,entities):
    """
    Order data from PRIMAP-crf into a 3D table (GHG x country x category).
    Uncovered categories/countries are set to 0

    Parameters
    ----------
    data : list of lists
        PRIMAP-crf base data (as returned by load_crf)
    countries : list of strings
        List of country A3 codes from EORA.
    cat_crf : list of strings
        List of emission categories.

    Returns
    -------
    output : 3D array
        Table of sourced emissions per GHG, country and emission category.
   """
    output = np.zeros([len(countries),len(entities),len(cat_crf)])
    for i in range(len(data[0])):
        country = countries.index(data[0][i][0])
        entity = entities.index(data[0][i][2])
        category = cat_crf.index(data[0][i][1])
        if data[1][i] != '':    
            output[country,entity,category] = data[1][i]
    return output

def order_primap_old(primap, data, known, countries,cat_crf,entities):
    """
    Include data from PRIMAP-hist into the estimated emission table.
    Include information on the data source.
    Primap_hist is given priority on pre-existing data.

    Parameters
    ----------
    primap : list of lists
        PRIMAP-hist data (as returned by load_hist)
    data : 3D matrix
        Estimated emission table (GHG x country x category)
    known : 3D matrix
        Miror of the emission table, holding information on data source.
    countries : list of strings
        List of country A3 codes from EORA.
    cat_crf : list of strings
        List of emission categories.

    Returns
    -------
    data : 3D array
        Updated table of emissions per GHG, country and emission category,
        including PRIMAP-hist data.
    known : 3D array
        Updated table of sources per data point.
        Data from PRIMAP-hist is marked as 2.
    """
    for index in range(len(primap[0])):
        country = countries.index(primap[0][index,0])
        category = cat_crf.index(primap[0][index,1])
        entity = entities.index(primap[0][index,2])
        if primap[1][index] != '' and primap[1][index] != 'NaN':
            data[country,entity,category] = primap[1][index]
            known[country,entity,category] = 2  
            #Mark available Primap data as certain
    return data,known

def aggregate(data,intensity,go,known,crf_parents,crf_childs,concordance):
    """
    Bottom up consolidation of the emission tree.
    Successively estimates emissions in the emission tree,
    starting from end-categories up to total emissions.
    Uses emission intensity coeffs when no data reported

    Parameters
    ----------
    data : 1D array
        IPCC emission data from PRIMAP-crf for a country/GHG.
    intensity : 1D array
        Intensity coefficient array for the corresponding GHG.
    go : 1D array
        Economic output per sector.
    known : 1D array
        Source status associated with source data.
            0 is unknown
            1 is estimated
            2 is known from PRIMAP-crf            
    crf_parents : list of ints
        Indices of parent categories.
    crf_childs : list of list of ints
        List of children categories per parent.
    concordance : list of ints
        Mapping of emission categories with economic sectors.

    Returns
    -------
    data : 1D array
        aggregated emission array.
    known : 1D array
        Updated information on data source.
    """
    aggregation = np.copy(data)
    #Aggregation step: bottom up estimation using the intensity factors
    for i in range(len(crf_parents),0,-1):
        #Identify parents and children categories
        parent = crf_parents[i-1]
        childs_id = crf_childs[i-1]
        for children in childs_id:
            if known[children] == 0 and concordance[children] !='':
                #If a children is null and is an end-category, 
                #estimate is computed using the intensity factor
                aggregation[children] = \
                    intensity[children]*go[int(concordance[children])]
                known[children] = 1 #Children is marked as aggregated
        if known[parent] == 0:
            #If the parent is unknown, child estimation are aggregated 
            #to form a parent estimation
            aggregation[parent] =\
                np.sum([aggregation[child] for child in childs_id])
            known[parent] = 1
    return aggregation,known

def down_adjustment(data,known,crf_parents,crf_childs):
    """
    Top-down adjustment of the emission tree.
    Ensures consistence between aggregated data and sub-categories.
    Adjusts estimated coefficients.
    If unconsistences in sourced data, known sub-categories are adjusted
    and estimated sub-categories are set to 0.
    
    Parameters
    ----------
    data : 1D numpy array
        Emissions values estimated for the emission categories 
        of the corresponding gas and country.
    known : 1D numpy array
        State of knowledge of the available information:
            0 is no information
            2 is original information
            1 is estimated value.
    crf_parents : list of ints
        list of aggregated emission categories.
    crf_childs : list of list of ints
        List of list of sub-categories associated with the parent categories.

    Returns
    -------
    data : 1D numpy array
        Updated emission values.
    known : 1D numpy array
        Updated information on the estimations:
            2 = source data
            x (float) = adjustment ratio
            -1 = has been set to 0 because children exceeded parents

    """
    for i in range(len(crf_parents)):   
        parent = crf_parents[i]
        childs = crf_childs[i]
        #When splitting data between children, known children left untouched
        sure = 0
        unsure = 0
        tobalance = []
        for child in childs:
            if known[child] == 2:
                sure += data[child] #Sums up the values of the sourced childs
            else:
                tobalance.append(child)
                unsure += data[child]   
                #Sums up the values of the aggregated childs
        tomake = data[parent] - sure 
        #Computes the emissions to allocate between the aggregated childs
        for kid in tobalance:
            if unsure > 0 and tomake > 0:  
                #If residual is positive, 
                #proportionally allocate extra emissions to aggregated children
                data[kid] = data[kid]*tomake/unsure
                known[kid] = tomake/unsure 
                #Support array marks the adjustment coefficient
            if tomake <= 0: 
                #If residual is negative, estimated emissions are assumed null
                data[kid] = 0
                #Support array marks that extrapolation is not possible 
                #because all emissions are already allocated
                known[kid] = -1 
        children = [data[children] for children in childs]
        made = np.sum(children)
        if made != data[parent]: 
            #If total of children are still different to parent, 
            #sourced childs are proportionally adjusted
            for children in childs:
                #If parent is reported but all children are null
                if made == 0:
                    #Emissions are allocated proportionally to childs
                    data[children] = data[parent]/len(childs)
                    known[children] = 1
                else:
                    data[children] = data[children]*data[parent]/made
                    known[children] = data[parent]/made 
                    #Support array registers the adjustment coefficient
        children = [data[children] for children in childs]
        assert np.isclose(np.sum(children),data[parent])
    return data, known

def kyoto(data,known,cat_crf,crf_parents,crf_childs,override=True):
    """
    Allocate Kyoto GHG basket, using AR4 Global Warming Potentials.
    If override is True, allocation is derived from allocation of other GHGs.
    Otherwise PRIMAP data is taken into account.

    Parameters
    ----------
    data : 3D array
        Emission table (GHG x Country x Category).
    known : 3D array
        Table mirroring data sources from emission data points.
    cat_crf : list of ints
        List of IPCC emission categories.
    crf_parents : list of ints
        list of aggregated emission categories.
    crf_childs : list of list of ints
        List of list of sub-categories associated with the parent categories.
    override : boolean, optional
        If True, data is for Kyoto basket is derived endogenously. 
        Otherwise, data is adjusted with regard to specific PRIMAP data.
        The default is True.

    Returns
    -------
    data : 3D array
        Updated emission table.
    known : 3D array
        Updated information on sources.

    """
    if override:
        data[4] = 25*data[0] + data[1] + data[2] + 298*data[3]
        known[4] = np.array([1]*len(known[4]))
    else:
        index=4
        for i in range(len(cat_crf)):
            if data[index,i] != 0:  
                #Mark available emission categories as known
                known[index,i] = 2
            else:
                #Estimate missing emission
                data[index,i] = \
                    25*data[0,i] + data[1,i] + data[2,i] + 298*data[3,i]
                known[index,i] = 1
        data[index],known[index] = down_adjustment(data[index],
                                                   known[index],
                                                   crf_parents,crf_childs)
    return data,known

def main(year,version,entities =['CO2']):
    """
    Maps PRIMAP data into EORA.
    Produces satellite matrices for GHGs listed in 'entities'.

    Parameters
    ----------
    year : int
        Base year to map.
    version : string
        Version for EORA data.

    Returns
    -------
    None. Files are saved directly in the output_path folder.

    """
    countries = load_countries(version)
    cat_crf,mapping,crf_parents,crf_childs = load_mapping()
    crf = load_crf(year, countries,cat_crf,entities)
    go = get_go(year,version)
    #Compute the intensity factors for each emission category and each GHG
    intensity,sources_intensity = intensity_factor(year,crf
                                                   ,mapping,cat_crf,
                                                   countries,go,entities)
    #Save file describing the input data for each category
    sources_intensity = pd.DataFrame(data=sources_intensity,
                                     index=entities,columns = cat_crf)
    #Order crf data into a 3D file
    #with information available for each GHG/country/category
    crf = order_primap(crf, countries,cat_crf,entities)
    #Create a coupled 'crf' file for storing the state of knowledge of the data
    computed = np.copy(crf)
    ratio = np.zeros([len(countries),len(entities),len(cat_crf)])
    ratio[crf!=0] = 2   #Mark known values as sourced inputs
    
    #First step: extrapolation using crf intensity factors
    for country_id in range(len(countries)):
        for entity_id in range(len(entities)):
            data = np.copy(crf[country_id,entity_id])    
            #Extract data for the corresponding country/gas from the crf table
            extract_i = intensity[entity_id]   
            #Extract data for the corresponding gas
            #from the intensity factors file
            goi = np.concatenate((go[0][country_id*26:(country_id+1)*26],
                                  [go[1][country_id]]))  
            #Extract and aggregate national gross output and final demand 
            #Incorporate estimations in 'computed' file
            #track changes (previously known/aggregated/unknown)
            #in the 'ratio' file
            data,ratio[country_id,entity_id] = \
                aggregate(data,extract_i,goi,ratio[country_id,entity_id],
                          crf_parents,crf_childs,mapping)
            computed[country_id,entity_id],ratio[country_id,entity_id] = \
                down_adjustment(data,ratio[country_id,entity_id],
                                crf_parents,crf_childs)
    test_consistency(computed,crf,ratio,crf_parents,crf_childs)
    
    ratio = np.zeros([len(countries),len(entities),len(cat_crf)])
    
    #Second step: integrate Primp-hist data
    for i,item in enumerate(cat_relabel[0]):
        cat_crf[cat_crf.index(item)] = cat_relabel[1][i]
    primap = load_hist(year, countries,cat_crf,entities)
    primap = order_primap(primap,countries,cat_crf,entities)
    computed[primap!=0] = primap[primap!=0]
    ratio[primap!=0] = 2
    for country_id in range(len(countries)):
        for entity_id in range(len(entities)):
            data = np.copy(primap[country_id,entity_id])
            #Set total emissions to 0 for countries not covered in primap_hist
            computed[country_id,entity_id,0] = primap[country_id,entity_id,0]
            computed[country_id,entity_id],ratio[country_id,entity_id] = \
                down_adjustment(computed[country_id,entity_id],
                                ratio[country_id,entity_id],
                                crf_parents,crf_childs)
    test_consistency(computed,primap,ratio,crf_parents,crf_childs)
    
    qt,qy = to_eora(computed,len(countries),
                    crf_parents,crf_childs,mapping,entities)
    go[0][go[0]==0] =1
    nqt = qt.dot(np.diag(1/go[0]))
    nqy = qy.dot(np.diag(1/go[1]))
    save(qt,qy,nqt,nqy,intensity,sources_intensity,version,year)

def to_eora(data,countries,crf_parents,crf_childs,mapping,
            entities,sectors=26,categories=[0]):
    """
    Maps emission tables into EORA sectors.

    Parameters
    ----------
    data : 3D array
        Emission table.
    countries : int
        Number of countries covered.
    crf_parents : list of ints
        list of aggregated emission categories.
    crf_childs : list of list of ints
        List of list of sub-categories associated with the parent categories.
    mapping : list of ints
        Concordance between EORA sectors and emission categories.
    entities : list of strings
        List of GHGs covered.
    sectors : int, optional
        Economic sectors per country The default is 26.
    categories : list of ints, optional
        Emission categories to report. The default is [0].

    Returns
    -------
    qt : 2D array
       Emission matrix for intermediate sectors.
    qy : 2D array
        Emission matrix for final demand.

    """
    n = len(categories)
    e = len(entities)
    qt = np.zeros((e*n,countries*sectors))
    qy = np.zeros((e*n,countries))
    
    
    selection = [look_for_children(i,crf_parents,crf_childs,[]) 
                 for i in categories]
    for entity in range(e):
        for country in range(countries):
            for i,cat in enumerate(selection):
                count = 0
                ref = data[country,entity,categories[i]]
                for item in cat:
                    count += data[country,entity,item]
                    sector = int(mapping[item])
                    if sector != 26:
                        qt[entity*n+i,country*26+sector] += data[country,
                                                                 entity,item] 
                    elif sector == 26:
                        qy[entity*n+i,country] += data[country,entity,item]
                assert np.isclose(count,ref)
    return qt,qy

def look_for_children(category,parents,childs,output=[]):
    '''
    Recursive function to find all end-categories to include
    in the reporting of selected category.

    Parameters
    ----------
    category : int
        Emission category under investigation
    parents : list of ints
        list of aggregated emission categories.
    childs : list of list of ints
        List of list of sub-categories associated with the parent categories.
    output : list of ints
        Aggregated list of categories covered by the inital category called.

    Returns
    -------
    output: list of ints
        List of end-categories covered by the initial category called.

    '''
    if category not in parents:
        output.append(category)
    else:
        for child in childs[parents.index(category)]:
            output=look_for_children(child,parents,childs,output)
    return output
        

def look_for_parent(categories,crf_parents,crf_childs,suspect):
    '''
    Recursive function to find all end-categories to include
    in the reporting of selected category.

    Parameters
    ----------
    categories : list of ints
        Emission categories to report separatly in the satellite matrix.
    crf_parents : list of ints
        list of aggregated emission categories.
    crf_childs : list of list of ints
        List of list of sub-categories associated with the parent categories.
    suspect : int
        Current category under examination.

    Returns
    -------
    TYPE
        List of end-categories to consider for each selected category.

    '''
    if suspect in categories:
        return suspect
    elif suspect == 0:
        return None
    else:
        parent = [line for line,childs in enumerate(crf_childs) 
                  if suspect in childs]
        return look_for_parent(categories,crf_parents,crf_childs,
                               crf_parents[parent[0]])

def test_consistency(test,source,known,crf_parents,crf_childs):
    """
    Test the consistency of the emission tree.
    Ensure that the sum of each children category equals its parent.
    Ensure that data from source is respected.

    Parameters
    ----------
    test : 3D array
        Emission table to assess.
    primap : 
    crf_parents : list of ints
        list of aggregated emission categories.
    crf_childs : list of list of ints
        List of list of sub-categories associated with the parent categories.

    Returns
    -------
    None. Asserts that sum of children equals parent for every pair.
    """
    for i in range(len(test)):
        for j in range(len(test[0])):
            assert all(test[i,j][known[i,j]==2] == source[i,j][known[i,j]==2])
            for k in range(len(crf_parents)):
                parent = test[i,j,crf_parents[k]]
                children = [test[i,j,child] for child in crf_childs[k]]
                children = np.sum(children)
                assert np.isclose(parent,children)

def save(qt,qy,nqt,nqy,intensity,sources,version,year,save_support=False):
    """
    Save satellite matrices in output folder.

    Parameters
    ----------
    qt : 2D array
        Emission of industries.
    qy : 2D array
        Emissions of final users.
    nqt : 2D array
        Intensity coefficients of industries.
    nqy : 2D array
        Intensity coefficients of final users.
    intensity : 2D array
        Average intensity coefficient per category.
    sources : 2D array
        Number of sources per average intensity coefficient.
    version : string
        EORA data version.
    year : int
        Year covered.
    save_support : boolean
        Whether to export the support matrices (intensity, sources,
        nqt, nqy). The defaukt is False

    Returns
    -------
    None.

    """
    np.savetxt(os.path.join(output_path,version,'QT_'+str(year)+'.csv'),
               qt,delimiter=',')
    np.savetxt(os.path.join(output_path,version,'QY_'+str(year)+'.csv'),
               qy,delimiter=',')
    if save_support:
        np.savetxt(os.path.join(output_path,version,'n_QT_'+str(year)+'.csv'),
               nqt,delimiter=',')
        np.savetxt(os.path.join(output_path,version,'n_QY_'+str(year)+'.csv'),
               nqy,delimiter=',')
        np.savetxt(os.path.join(output_path,version,
                            'intensity_sources_'+str(year)+'.csv'),
               sources,delimiter=',')
        np.savetxt(os.path.join(output_path,version,
                            'intensity_factors_'+str(year)+'.csv'),
               intensity,delimiter=',')

def every_year():
    """
    Compute the allocation for all years between 1990 and 2015
    Save them in Eora format as csv files
    """
    for year in range(1990,2016):
        main(year,'formatted')
        
main(2015,'formatted',['CO2'])